import 'package:sqflite/sqflite.dart';
import 'package:sql_query/builder/query.dart';

import '../database_model.dart';

class DBMerge {
  List<TableDatabase> tableList;
  Database? database;
  Function? onInitComplete;
  DBMerge({
    required this.database,
    required this.tableList,
    this.onInitComplete,
  }) {
    _init();
  }

  Query table(String table) {
    return Query(
      table: table,
      database: this,
    );
  }

  _init() async {
    List<String> insertList = [];
    List<TableDatabase> foreignKey = [];
    for (final tl in tableList) {
      insertList.add(_tableBuilder(tl));
    }

    for (final tl in foreignKey) {
      insertList.add(_tableBuilder(tl));
    }

    await database!.transaction((txn) async {
      final list = await txn.rawQuery('SELECT * FROM sqlite_master');
      for (var i in insertList) {
        if (list.where((element) => element['sql'] == i).isEmpty) {
          try {
            await txn.execute(i);
          } catch (_) {
            // await txn.execute(i.replaceAll('CREATE', 'ALTER'));
          }
        }
      }
    });
    if (onInitComplete != null) {
      onInitComplete!();
    }
  }

  String _tableBuilder(TableDatabase data) {
    String insert = 'CREATE TABLE ${data.tableName} (';
    int total = 0;
    for (final tc in data.column) {
      ++total;
      String prop = '';

      if (tc.typeData == DataType.text) {
        prop += ' TEXT';
      } else if (tc.typeData == DataType.integer) {
        // if (tc.typeData == DataType.bool) {
        //   _boolData.add(tc.name);
        // }
        prop += ' INTEGER';
      } else if (tc.typeData == DataType.double) {
        prop += ' REAL';
      }

      if (tc.primaryKey) {
        prop += ' PRIMARY KEY';
      }

      if (tc.autoIncrement) {
        prop += ' AUTOINCREMENT';
      }

      if (!tc.isNull) {
        prop += ' NOT NULL';
      }

      if (tc.defaultValue != null) {
        prop += ' DEFAULT ${tc.typeData == DataType.text ? '"' : ''}${tc.defaultValue}${tc.typeData == DataType.text ? '"' : ''}';
      }

      insert += '${tc.name} ' + prop;
      if (data.column.length > 1 && total != data.column.length) {
        insert += ', ';
      }
    }
    if (data.foreignKey != null) {
      insert += ' , ';
      insert += ' FOREIGN KEY(${data.foreignKey?.baseColumn}) REFERENCES ${data.foreignKey?.referenceTable}(${data.foreignKey?.referenceColumn}) ';
      if (data.foreignKey!.deleteAction != null) {
        insert += ForeignKey.getAction(data.foreignKey!.deleteAction!, 'DELETE');
      }

      if (data.foreignKey!.updateAction != null) {
        insert += ForeignKey.getAction(data.foreignKey!.updateAction!, 'UPDATE');
      }
    }
    insert += ')';
    return insert;
  }

  Future<bool> insert({dynamic data, tableName}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          if (data is List) {
            for (Map<String, String> i in data) {
              final Map<String, String> maps = {};
              i.forEach((key, value) {
                final parse = value.replaceAll("'", "|_|_|").replaceAll('"', '|-|-|');
                maps.addAll({key: parse});
              });
              batch.insert(tableName, maps);
            }
          } else {
            batch.insert(tableName, data);
          }
          await batch.commit(continueOnError: false);
          success = true;
        } catch (_) {
          print(_);
          success = false;
        }
      },
    );
    return success;
  }

  Future<bool> update({dynamic data, where, tableName}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          if (data is List) {
            for (Map<String, String> i in data) {
              final Map<String, String> maps = {};
              i.forEach((key, value) {
                maps.addAll({key: value.replaceAll("'", "|_|_|").replaceAll('"', '|-|-|')});
              });
              batch.update(tableName, maps, where: where);
            }
          } else {
            batch.update(tableName, data, where: where);
          }
          await batch.commit(continueOnError: false);
          success = true;
        } catch (_) {
          success = false;
          print(_);
        }
      },
    );
    return success;
  }

  Future<bool> delete({query}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          batch.rawQuery(query);
          await batch.commit(continueOnError: false);
          success = true;
        } catch (_) {
          success = false;
          print(_);
        }
      },
    );
    return success;
  }

  Future<List<Map<String, dynamic>>> get({
    query,
  }) async {
    List<Map<String, dynamic>> list = await database!.rawQuery(
      query,
    );
    List<Map<String, dynamic>> lists = [];
    for (var i in list) {
      Map<String, dynamic> map = {};
      i.forEach((key, value) {
        if (value is String) {
          value = value.replaceAll('|_|_|', "'").replaceAll('|-|-|', '"');
          map.addAll({key: value});
        } else {
          map.addAll({key: value});
        }
      });
      lists.add(map);
    }
    return lists;
  }
}
