import '../database_config.dart';
import '../random.dart';
import 'db_merge.dart';

class Query {
  String table;
  DBMerge? database;
  String _lastQuery = '';
  String _joinQuery = '';
  String _groupBy = '';
  String _orderBy = '';
  String _limit = '';
  String _select = '';
  bool _self = false;
  final List<Map<String, String>> _data = [];

  Query({required this.table, this.database});

  Query listData(List<Map<String, String>> listData) {
    _data.insertAll(_data.length, listData);
    return this;
  }

  Query data(Map<String, String> data) {
    _data.add(data);
    return this;
  }

  Query limit(int limit) {
    if (!_joinQuery.contains('LIMIT')) {
      _limit += ' LIMIT $limit ';
    }
    return this;
  }

  /// ```[ table1.id DESC , table2.id ASC ]```
  Query orderBy(List<String> columnName) {
    final order = columnName.toString().replaceAll('[', '').replaceAll(']', '');
    _orderBy += ' ORDER BY $order ';
    return this;
  }

  Query groupBy(List<String> columnName) {
    final groupB = columnName.toString().replaceAll('[', '').replaceAll(']', '');
    if (groupB != 'null') {
      _groupBy += ' GROUP BY $groupB';
    }
    return this;
  }

  Query select(List<String> column) {
    if (_select != '') {
      _select += ', ';
    }
    _select += column.toString().replaceAll('[', '').replaceAll(']', '');
    return this;
  }

  Query where(column, operator, value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }

    _lastQuery += (column) + ' ' + (operator ?? '=') + ' ' + (value == null ? 'NULL' : checkIntType(value));
    return this;
  }

  Query orWhereNotIn(column, List<String> value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    int count = 0;
    _lastQuery += ' ( ';
    for (var i in value) {
      count++;
      if (count > 1) {
        _lastQuery += ' AND ';
      }
      _lastQuery += (column) + ' ' + '!=' + ' ' + (checkIntType(i)) + ' ';
    }
    _lastQuery += ' ) ';
    return this;
  }

  Query orWhereTableNotIn(column, String getQueryTable) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    _lastQuery += ' $column NOT IN ( $getQueryTable ) ';
    return this;
  }

  Query whereNotIn(column, List<String> value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    _lastQuery += ' ( ';
    int count = 0;
    for (var i in value) {
      count++;
      if (count > 1) {
        _lastQuery += ' AND ';
      }
      _lastQuery += (column) + ' ' + '!=' + ' ' + (checkIntType(i)) + ' ';
    }
    _lastQuery += ' ) ';
    return this;
  }

  Query whereTableNotIn(column, String getQueryTable) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    _lastQuery += ' $column NOT IN ( $getQueryTable ) ';
    return this;
  }

  Query orWhereIn(column, List<String> value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    int count = 0;
    _lastQuery += ' ( ';
    for (var i in value) {
      count++;
      if (count > 1) {
        _lastQuery += ' AND ';
      }
      _lastQuery += (column) + ' ' + '=' + ' ' + (checkIntType(i)) + ' ';
    }
    _lastQuery += ' ) ';
    return this;
  }

  Query orWhereTableIn(column, String getQueryTable) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    _lastQuery += ' IN ( $getQueryTable ) ';
    return this;
  }

  Query whereIn(column, List<String> value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    int count = 0;
    _lastQuery += ' ( ';
    for (var i in value) {
      count++;
      if (count > 1) {
        _lastQuery += ' AND ';
      }
      _lastQuery += (column) + ' ' + '=' + ' ' + (checkIntType(i)) + ' ';
    }
    _lastQuery += ' ) ';
    return this;
  }

  Query whereTableIn(column, String getQueryTable) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    _lastQuery += ' IN ( $getQueryTable ) ';
    return this;
  }

  Query whereBetween(columnName, startValue, endValue) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }

    _lastQuery += ' $columnName BETWEEN ';

    final tryParse = [DateTime.tryParse(startValue), DateTime.tryParse(endValue)];
    int count = 0;
    for (var i in tryParse) {
      ++count;
      if (count == 2) {
        _lastQuery += ' AND ';
      }
      if (i != null) {
        _lastQuery += i.millisecondsSinceEpoch.toString() + ' ';
      } else {
        int? tryInt;
        if (count == 1) {
          tryInt = int.tryParse(startValue);
        } else if (count == 2) {
          tryInt = int.tryParse(endValue);
        }

        if (tryInt != null) {
          _lastQuery += tryInt.toString() + ' ';
        }
      }
    }

    return this;
  }

  Query orWhereBetween(columnName, startValue, endValue) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }

    _lastQuery += ' $columnName BETWEEN ';

    final tryParse = [DateTime.tryParse(startValue), DateTime.tryParse(endValue)];
    int count = 0;
    for (var i in tryParse) {
      ++count;

      if (count == 2) {
        _lastQuery += ' AND ';
      }
      if (i != null) {
        _lastQuery += i.millisecondsSinceEpoch.toString() + ' ';
      } else {
        int? tryInt;
        if (count == 1) {
          tryInt = int.tryParse(startValue);
        } else if (count == 2) {
          tryInt = int.tryParse(endValue);
        }

        if (tryInt != null) {
          _lastQuery += tryInt.toString() + ' ';
        }
      }
    }

    return this;
  }

  Query orWhere(column, operator, value) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }

    _lastQuery += (column) + ' ' + (operator ?? '=') + ' ' + (checkIntType(value));
    return this;
  }

  Query whereHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    String has = ' EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query orWhereHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    String has = ' EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query whereNotHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    String has = ' NOT EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query orWhereNotHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    String has = ' NOT EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query on(tableColumn, baseColumn) {
    if (_joinQuery.contains('{TABLE_COLUMN_FILL}')) {
      _joinQuery += ' AND ';
    }
    if (_self) {
      _joinQuery += table + '.' + tableColumn;
      _joinQuery += '=';
      _joinQuery += '{TABLE_COLUMN_FILL}.' + baseColumn;
    } else {
      _joinQuery += '{TABLE_COLUMN_FILL}.' + tableColumn + '=' + table + '.' + baseColumn;
    }
    return this;
  }

  Query join(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += 'INNER JOIN  $tableName ON ';

      Query q = function(Query(table: baseTableCustom ?? table));
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      _joinQuery += ' INNER JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query leftJoin(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += 'LEFT JOIN  $tableName ON ';

      Query q = function(Query(table: baseTableCustom ?? table));
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      _joinQuery += ' LEFT JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query crossJoin(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += ' ( ';
      _joinQuery += 'CROSS JOIN  $tableName ON ';
      Query q = function(Query(table: baseTableCustom ?? table));
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      _joinQuery += ' CROSS JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query selfJoin(tableName, {columnName, baseColumnName, Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += 'INNER JOIN $tableName $rand ON ';
      Query q = function(Query(table: tableName), tableName, rand);
      q._self = false;
      q.table = table;
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' INNER JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + columnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + baseColumnName;
    }
    return this;
  }

  Query selfLeftJoin(tableName, {baseColumnName, columnName, Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += 'LEFT JOIN $tableName $rand ON ';
      final que = Query(table: tableName);
      que.table = tableName;
      que._self = true;
      Query q = function(que, tableName, rand);
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' LEFT JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + columnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + baseColumnName;
    }
    return this;
  }

  Query selfCrossJoin(tableName, {baseColumnName, columnName, Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += 'CROSS JOIN $tableName $rand ON ';
      final que = Query(table: tableName);
      que.table = tableName;
      que._self = true;
      Query q = function(que, tableName, rand);
      _select += (_select != '' ? ',' : '') + q._select;
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' CROSS JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + columnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + baseColumnName;
    }
    return this;
  }

  Future<List<Map<String, dynamic>>> insert() async {
    List<Map<String, dynamic>> result = [];
    final res = database != null ? await database!.insert(data: _data, tableName: table) : await DatabaseConfig.insert(data: _data, tableName: table);
    if (res) {
      result.add({'status': 'success', 'message': 'data has been Added'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> update() async {
    List<Map<String, dynamic>> result = [];
    final res = database != null ? await database!.update(data: _data, tableName: table, where: _lastQuery.replaceAll('WHERE', '')) : await DatabaseConfig.update(data: _data, tableName: table, where: _lastQuery.replaceAll('WHERE', ''));
    if (res) {
      result.add({'status': 'success', 'message': 'data has been Updated'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> delete() async {
    List<Map<String, dynamic>> result = [];
    final lastQ = 'DELETE FROM $table ' + _joinQuery + ' ' + _lastQuery + ' ' + _groupBy + ' ' + _orderBy + ' ' + _limit;
    final res = database != null ? await database!.delete(query: lastQ) : await DatabaseConfig.delete(query: lastQ);
    if (res) {
      result.add({'status': 'success', 'message': 'data has been deleted'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> get() async {
    final lastQ = 'SELECT ${_select == '' ? '*' : _select} FROM $table ' + _joinQuery + ' ' + _lastQuery + ' ' + _groupBy + ' ' + _orderBy + ' ' + _limit;
    return database != null ? await database!.get(query: lastQ) : await DatabaseConfig.get(query: lastQ);
  }

  String getQuery() {
    final lastQ = 'SELECT ${_select == '' ? '*' : _select} FROM $table ' + _joinQuery + ' ' + _lastQuery + ' ' + _groupBy + ' ' + _orderBy + ' ' + _limit;
    return lastQ;
  }
}
