import 'package:sql_query/builder/query.dart';

import '../database_config.dart';
import '../database_model.dart';

class DB {
  static Query table(String table) {
    return Query(
      table: table,
    );
  }

  static String _tableBuilder(TableDatabase data) {
    String insert = 'CREATE TABLE ${data.tableName} (';
    int total = 0;
    for (final tc in data.column) {
      ++total;
      String prop = '';

      if (tc.typeData == DataType.text) {
        prop += ' TEXT';
      } else if (tc.typeData == DataType.integer) {
        // if (tc.typeData == DataType.bool) {
        //   _boolData.add(tc.name);
        // }
        prop += ' INTEGER';
      } else if (tc.typeData == DataType.double) {
        prop += ' REAL';
      }

      if (tc.primaryKey) {
        prop += ' PRIMARY KEY';
      }

      if (tc.autoIncrement) {
        prop += ' AUTOINCREMENT';
      }

      if (!tc.isNull) {
        prop += ' NOT NULL';
      }

      if (tc.defaultValue != null) {
        prop += ' DEFAULT ${tc.typeData == DataType.text ? '"' : ''}${tc.defaultValue}${tc.typeData == DataType.text ? '"' : ''}';
      }

      insert += '${tc.name} ' + prop;
      if (data.column.length > 1 && total != data.column.length) {
        insert += ', ';
      }
    }
    if (data.foreignKey != null) {
      insert += ' , ';
      insert += ' FOREIGN KEY(${data.foreignKey?.baseColumn}) REFERENCES ${data.foreignKey?.referenceTable}(${data.foreignKey?.referenceColumn}) ';
      if (data.foreignKey!.deleteAction != null) {
        insert += ForeignKey.getAction(data.foreignKey!.deleteAction!, 'DELETE');
      }

      if (data.foreignKey!.updateAction != null) {
        insert += ForeignKey.getAction(data.foreignKey!.updateAction!, 'UPDATE');
      }
    }
    insert += ')';
    return insert;
  }

  static init({required List<TableDatabase> tableList, String? databaseName}) async {
    List<String> insertList = [];
    List<TableDatabase> foreignKey = [];
    for (final tl in tableList) {
      insertList.add(_tableBuilder(tl));
    }

    for (final tl in foreignKey) {
      insertList.add(_tableBuilder(tl));
    }

    await DatabaseConfig.init(tables: insertList, dbName: databaseName);
  }

  static deleteDB() {
    DatabaseConfig.dbDelete();
  }

  static closeDB() {
    DatabaseConfig.dbClose();
  }
}
